#!/usr/bin/env python
# (c) Copyright 2010 Cloudera, Inc. All rights reserved.
#
# Handling of data export

import cStringIO
import csv
import logging
import pyExcelerator as xl
import time

from django.http import HttpResponse

from beeswax import common
from beeswax import db_utils
from beeswaxd.ttypes import QueryHandle

XLS_SIZE_LIMIT = 200 * 1024 * 1024      # 200MB

LOG = logging.getLogger(__name__)

_DATA_WAIT_SLEEP = 0.1                  # Sleep 0.1 sec before checking for data availability


class TooBigToDownloadException(Exception):
  pass


def download(query_model, format):
  """
  download(query_model, format) -> HttpResponse

  Retrieve the query result in the format specified. Return an HttpResponse object.
  """
  if format not in common.DL_FORMATS:
    LOG.error('Unknown download format "%s"' % (format,))
    return

  if format == 'csv':
    formatter = CSVformatter()
    mimetype = 'application/csv'
  elif format == 'xls':
    formatter = XLSformatter()
    mimetype = 'application/xls'

  gen = data_generator(query_model, formatter)
  resp = HttpResponse(gen, mimetype=mimetype)
  resp['Content-Disposition'] = 'attachment; filename=query_result.%s' % (format,)
  return resp


def data_generator(query_model, formatter):
  """
  data_generator(query_model, formatter) -> generator object

  Return a generator object for a csv. The first line is the column names.
  """
  global _DATA_WAIT_SLEEP
  is_first_row = True
  next_row = 0
  results = None
  handle = QueryHandle(query_model.server_id, query_model.log_context)

  yield formatter.init_doc()

  while True:
    # Make sure that we have the next batch of ready results
    while results is None or not results.ready:
      results = db_utils.db_client().fetch(handle, start_over=is_first_row)
      if not results.ready:
        time.sleep(_DATA_WAIT_SLEEP)

    # Someone is reading the results concurrently. Abort.
    # But unfortunately, this current generator will produce incomplete data.
    if next_row != results.start_row:
      msg = 'Error: Potentially incomplete results as an error occur during data retrieval.'
      yield formatter.format_row([msg])
      err = ('Detected another client retrieving results for %s. '
             'Expect next row being %s and got %s. Aborting' %
             (query_model.server_id, next_row, results.start_row))
      LOG.error(err)
      raise RuntimeError(err)

    if is_first_row:
      is_first_row = False
      yield formatter.format_header(results.columns)
    else:
      for i, row in enumerate(results.data):
        # TODO(bc): Hive seems to always return tab delimited row data.
        # What if a cell has a tab?
        row = row.split('\t')
        try:
          yield formatter.format_row(row)
        except TooBigToDownloadException, ex:
          LOG.error(ex)
          # Exceeded limit. Stop.
          results.has_more = False
          break

      if results.has_more:
        next_row += len(results.data)
        results = None
      else:
        yield formatter.fini_doc()
        break


class Formatter(object):
  """
  The interface for a helper class to write formatted data.
  """
  def init_doc(self):
    """
    init_doc() -> initial data to appear before the header
    """
    raise NotImplementedError

  def format_header(self, header):
    """
    format_header(header) -> line
    header should be a list of column names
    """
    raise NotImplementedError

  def format_row(self, row):
    """
    format_row(row) -> line
    row should be a list of datum

    May raise TooBigToDownloadException. In which case fini_doc() will still be
    invoked to make the partial results available to the user. Implementation may
    want to signify that the results are partial in the doc.
    """
    raise NotImplementedError

  def fini_doc(self):
    """
    fini_doc() -> final data to appear after all rows
    """
    raise NotImplementedError


class CSVformatter(Formatter):
  def __init__(self):
    dialect = csv.excel()
    dialect.quoting = csv.QUOTE_ALL
    self._csv_writer = csv.writer(self, dialect=dialect)
    self._line = None

  def write(self, str):
    self._line = str

  def init_doc(self):
    return ""

  def format_header(self, header):
    return self.format_row(header)

  def format_row(self, row):
    # writerow will call our write() method
    self._csv_writer.writerow(row)
    return self._line

  def fini_doc(self):
    return ""


class XLSformatter(Formatter):
  """Unfortunately, pyExcelerator can't stream."""
  def __init__(self):
    self._book = xl.Workbook()
    self._sheet = self._book.add_sheet("Sheet 1")
    self._row = 0
    self._size = 0

  def init_doc(self):
    return ''

  def format_header(self, header):
    for i, cell in enumerate(header):
      self._sheet.write(self._row, i, cell)
    self._row += 1
    return ''

  def format_row(self, row):
    for i, cell in enumerate(row):
      self._sheet.write(self._row, i, cell)
      self._limit_size(cell)
    self._row += 1
    return ''

  def fini_doc(self):
    string_io = cStringIO.StringIO()
    self._book.save(string_io)
    res = string_io.getvalue()
    string_io.close()
    return res

  def _limit_size(self, cell):
    self._size += len(cell)
    if self._size > XLS_SIZE_LIMIT:
      msg = 'ERROR: Data too large to be downloaded as XLS file.'
      self._row += 1
      self._sheet.write(self._row, 0, msg)
      raise TooBigToDownloadException(msg)
