#!/usr/bin/env python
# (c) Copyright 2010 Cloudera, Inc. All rights reserved.

from django import forms
from desktop.lib.django_forms import simple_formset_factory, DependencyAwareForm
from desktop.lib.django_forms import ChoiceOrOtherField, MultiForm, SubmitButton
from beeswax import common
from beeswax import models

import filebrowser.forms

def query_form():
  """Generates a multi form object for queries."""
  return MultiForm(
      query=HQLForm,
      settings=SettingFormSet,
      file_resources=FileResourceFormSet,
      functions=FunctionFormSet,
      saveform=SaveForm)


class SaveForm(forms.Form):
  """Used for saving query design and report design."""
  name = forms.CharField(required=False,
                        max_length=64,
                        initial=models.SavedQuery.DEFAULT_NEW_DESIGN_NAME,
                        help_text='Change the name to save as a new design')
  desc = forms.CharField(required=False, max_length=1024, label="Description")
  save = forms.BooleanField(widget=SubmitButton, required=False)
  saveas = forms.BooleanField(widget=SubmitButton, required=False)

  def __init__(self, *args, **kwargs):
    forms.Form.__init__(self, *args, **kwargs)
    self.fields['save'].label = 'Save'
    self.fields['save'].widget.label = 'Save'
    self.fields['saveas'].label = 'Save As'
    self.fields['saveas'].widget.label = 'Save As'

  def clean_name(self):
    name = self.cleaned_data.get('name', '').strip()
    return name

  def clean(self):
    if self.errors:
      return
    save = self.cleaned_data.get('save')
    name = self.cleaned_data.get('name')
    if save and len(name) == 0:
      # Bother with name iff we're saving
      raise forms.ValidationError('Please enter a name')
    return self.cleaned_data

  def set_data(self, name, desc=None):
    """Set the name and desc programmatically"""
    data2 = self.data.copy()
    data2[self.add_prefix('name')] = name
    if desc:
      data2[self.add_prefix('desc')] = desc
    self.data = data2


class HQLForm(forms.Form):
  query = forms.CharField(label="Query Editor", required=True, widget=forms.Textarea(attrs={'class':'beeswax_query'}))

class FunctionForm(forms.Form):
  name = forms.CharField(required=True)
  class_name = forms.CharField(required=True)

FunctionFormSet = simple_formset_factory(FunctionForm)


class FileResourceForm(forms.Form):
  type = forms.ChoiceField(required=True,
    choices=[
      ("JAR", "jar"),
      ("ARCHIVE", "archive"),
      ("FILE", "file"),
    ], help_text="Resources to upload with your Hive job." +
       "  Use 'jar' for UDFs.  Use file and archive for "
       "side files and MAP/TRANSFORM using.  Paths are on HDFS."
  )
  # TODO(philip): Could upload files here, too.  Or merely link
  # to upload utility?
  path = forms.CharField(required=True, help_text="Path to file on HDFS.")

FileResourceFormSet = simple_formset_factory(FileResourceForm)


class SettingForm(forms.Form):
  # TODO: There are common settings that should be always exposed?
  key = forms.CharField()
  value = forms.CharField()

SettingFormSet = simple_formset_factory(SettingForm)


# In theory, there are only 256 of these...
TERMINATOR_CHOICES = [
  (r'\001', r"'^A' (\001)"),
  (r'\002', r"'^B' (\002)"),
  (r'\003', r"'^C' (\003)"),
  (r'\t', r"Tab (\t)"),
  (',', "Comma (,)"),
  (" ", "Space"),
]

class CreateTableForm(DependencyAwareForm):
  """
  Form used in the create table page
  """
  dependencies = []

  # Basic Data
  name = common.HiveIdentifierField(label="Table Name", required=True)
  comment = forms.CharField(label="Description", required=False)

  # Row Formatting
  row_format = forms.ChoiceField(required=True,
                                choices=common.to_choices([ "Delimited", "SerDe" ]),
                                initial="Delimited")

  # Delimited Row
  # These initials are per LazySimpleSerDe.DefaultSeparators
  field_terminator = ChoiceOrOtherField(required=False, initial=TERMINATOR_CHOICES[0][0],
    choices=TERMINATOR_CHOICES)
  collection_terminator = ChoiceOrOtherField(required=False, initial=TERMINATOR_CHOICES[1][0],
    choices=TERMINATOR_CHOICES)
  map_key_terminator = ChoiceOrOtherField(required=False, initial=TERMINATOR_CHOICES[2][0],
    choices=TERMINATOR_CHOICES)
  dependencies += [
    ("row_format", "Delimited", "field_terminator"),
    ("row_format", "Delimited", "collection_terminator"),
    ("row_format", "Delimited", "map_key_terminator"),
  ]

  # Serde Row
  serde_name = forms.CharField(required=False, label="SerDe Name")
  serde_properties = forms.CharField(
                        required=False,
                        help_text="Comma-separated list of key-value pairs, eg., 'p1=v1, p2=v2'")

  dependencies += [
    ("row_format", "SerDe", "serde_name"),
    ("row_format", "SerDe", "serde_properties"),
  ]

  # File Format
  file_format = forms.ChoiceField(required=False, initial="TextFile",
                        choices=common.to_choices(["TextFile", "SequenceFile", "InputFormat"]),
                        widget=forms.RadioSelect)
  input_format_class = forms.CharField(required=False, label="InputFormat Class")
  output_format_class = forms.CharField(required=False, label="OutputFormat Class")

  dependencies += [
    ("file_format", "InputFormat", "input_format_class"),
    ("file_format", "InputFormat", "output_format_class"),
  ]

  # External?
  use_default_location = forms.BooleanField(required=False, initial=True,
    label="Use default location")
  external_location = forms.CharField(required=False, help_text="Path to HDFS directory or file of table data.")

  dependencies += [
    ("use_default_location", False, "external_location")
  ]


# Note, struct is not currently supported.  (Because it's recursive, for example.)
HIVE_TYPES = \
    ( "string", "tinyint", "smallint", "int", "bigint", "boolean",
      "float", "double", "array", "map",)
HIVE_PRIMITIVE_TYPES = \
    ("string", "tinyint", "smallint", "int", "bigint", "boolean", "float", "double")

class PartitionTypeForm(forms.Form):
  column_name = common.HiveIdentifierField(required=True)
  column_type = forms.ChoiceField(required=True, choices=common.to_choices(HIVE_PRIMITIVE_TYPES))

class ColumnTypeForm(DependencyAwareForm):
  """
  Form used to specify a column during table creation
  """
  dependencies = [
    ("column_type", "array", "array_type"),
    ("column_type", "map", "map_key_type"),
    ("column_type", "map", "map_value_type"),
  ]
  column_name = common.HiveIdentifierField(required=True)
  column_type = forms.ChoiceField(required=True, 
    choices=common.to_choices(HIVE_TYPES))
  array_type = forms.ChoiceField(required=False, 
    choices=common.to_choices(HIVE_PRIMITIVE_TYPES), label="Array Value Type")
  map_key_type = forms.ChoiceField(required=False,
                                   choices=common.to_choices(HIVE_PRIMITIVE_TYPES),
                                   help_text="Specify if column_type is map.")
  map_value_type = forms.ChoiceField(required=False,
                                     choices=common.to_choices(HIVE_PRIMITIVE_TYPES),
                                     help_text="Specify if column_type is map.")

ColumnTypeFormSet = simple_formset_factory(ColumnTypeForm, initial=[{}], add_label="add a column")
# Default to no partitions
PartitionTypeFormSet = simple_formset_factory(PartitionTypeForm, add_label="add a partition")


class LoadDataForm(forms.Form):
  """Form used for loading data into an existing table."""
  path = filebrowser.forms.PathField(label="Path")
  overwrite = forms.BooleanField(required=False, initial=False,
    label="Overwrite?")

  def __init__(self, table_obj, *args, **kwargs):
    """
    @param table_obj is a hive_metastore.thrift Table object,
    used to add fields corresopnding to partition keys.
    """
    super(LoadDataForm, self).__init__(*args, **kwargs)
    self.partition_columns = dict()
    for i, column in enumerate(table_obj.partitionKeys):
      # We give these numeric names because column names
      # may be unpleasantly arbitrary.
      name = "partition_%d" % i
      char_field = forms.CharField(required=True, label="%s (partition key with type %s)" % (column.name, column.type))
      self.fields[name] = char_field
      self.partition_columns[name] = column.name
