#!/usr/bin/env python
# (c) Copyright 2010 Cloudera, Inc. All rights reserved.
#

"""
Common infrastructure for beeswax tests
"""

import atexit
import logging
import pwd
import os
import subprocess
import time

import fb303.ttypes
from nose.tools import assert_true, assert_false

from desktop.lib.django_test_util import make_logged_in_client

from hadoop import mini_cluster
import hadoop.conf

import beeswax.conf


_INITIALIZED = False
_SHARED_BEESWAX_SERVER_PROCESS = None


BEESWAXD_TEST_PORT = 6969
LOG = logging.getLogger(__name__)

def _start_server(cluster):
  """
  Start beeswaxd and metastore
  """
  script = beeswax.conf.BEESWAX_SERVER_BIN.get()
  args = [
    script,
    '--beeswax',
    str(BEESWAXD_TEST_PORT),
    '--metastore',
    str(BEESWAXD_TEST_PORT + 1),
    '--superuser',
    pwd.getpwuid(os.getuid())[0]
  ]
  env = {
    'HADOOP_HOME': hadoop.conf.HADOOP_HOME.get(),
    'HADOOP_CONF_DIR': cluster.config_dir,
    'HIVE_CONF_DIR': beeswax.conf.BEESWAX_HIVE_CONF_DIR.get(),
  }
  if os.getenv("JAVA_HOME"):
    env["JAVA_HOME"] = os.getenv("JAVA_HOME")

  LOG.info("Executing %s, env %s, cwd %s" % (repr(args), repr(env), cluster.tmpdir))
  process = subprocess.Popen(args=args, env=env, cwd=cluster.tmpdir, stdin=subprocess.PIPE)
  return process



def get_shared_beeswax_server():
  finish = (
    beeswax.conf.BEESWAX_SERVER_HOST.set_for_testing("localhost"),
    beeswax.conf.BEESWAX_SERVER_PORT.set_for_testing(BEESWAXD_TEST_PORT),
    beeswax.conf.BEESWAX_META_SERVER_HOST.set_for_testing("localhost"),
    beeswax.conf.BEESWAX_META_SERVER_PORT.set_for_testing(BEESWAXD_TEST_PORT + 1),
    # Use a bogus path to avoid loading the normal hive-site.xml
    beeswax.conf.BEESWAX_HIVE_CONF_DIR.set_for_testing('/my/bogus/path'),
  )

  cluster = mini_cluster.shared_cluster(conf=True)

  global _SHARED_BEESWAX_SERVER_PROCESS
  if _SHARED_BEESWAX_SERVER_PROCESS is None:
    p = _start_server(cluster)
    _SHARED_BEESWAX_SERVER_PROCESS = p
    def kill():
      LOG.info("Killing beeswax server (pid %d)." % p.pid)
      os.kill(p.pid, 9)
      p.wait()
    atexit.register(kill)
    # Wait for server to come up, by repeatedly trying.
    start = time.time()
    started = False
    sleep = 0.001
    while not started and time.time() - start < 20.0:
      try:
        client = beeswax.db_utils.db_client()
        meta_client = beeswax.db_utils.meta_client()

        client.echo("echo")
        if meta_client.getStatus() == fb303.ttypes.fb_status.ALIVE:
          started = True
          break
        time.sleep(sleep)
        sleep *= 2
      except:
        pass
    if not started:
      raise Exception("Beeswax server took too long to come up.")

  def s():
    for f in finish:
      f()
    cluster.shutdown()

  return cluster, s


def wait_for_query_to_finish(client, response, max=30.0):
  logging.info(str(response.template) + ": " + str(response.content))
  start = time.time()
  sleep_time = 0.05
  # We don't check response.template == "watch_wait.mako" here,
  # because Django's response.template stuff is not thread-safe.
  while "Waiting for query..." in response.content:
    time.sleep(sleep_time)
    sleep_time = min(1.0, sleep_time * 2) # Capped exponential
    if (time.time() - start) > max:
      LOG.warning("Query took too long!")
      raise Exception("Query took too long.")
    response = client.get(response.request["PATH_INFO"], follow=True)
  return response


def make_query(client, query, submission_type="Execute",
               follow=True, udfs=None, settings=None, resources=[],
               wait=False, name=None, desc=None, local=True):
  """
  Prepares arguments for the execute view.

  If wait is True, waits for query to finish as well.
  """

  if settings is None:
    settings = []
  if local:
    # Tests run faster if not run against the real cluster.
    settings.append( ("mapred.job.tracker", "local") )

  # Prepares arguments for the execute view.
  parameters = {
    'query-query': query,
  }
  if submission_type == 'Execute':
    parameters['button-submit'] = 'Whatever'
  elif submission_type == 'Explain':
    parameters['button-explain'] = 'Whatever'
  elif submission_type == 'Save':
    parameters['saveform-save'] = 'True'

  if name:
    parameters['saveform-name'] = name
  if desc:
    parameters['saveform-desc'] = desc

  parameters["functions-next_form_id"] = str(len(udfs or []))
  for i, udf_pair in enumerate(udfs or []):
    name, klass = udf_pair
    parameters["functions-%d-name" % i] = name
    parameters["functions-%d-class_name" % i] = klass
    parameters["functions-%d-_exists" % i] = 'True'
  parameters["settings-next_form_id"] = str(len(settings))
  for i, settings_pair in enumerate(settings or []):
    key, value = settings_pair
    parameters["settings-%d-key" % i] = key
    parameters["settings-%d-value" % i] = value
    parameters["settings-%d-_exists" % i] = 'True'
  parameters["file_resources-next_form_id"] = str(len(resources or []))
  for i, resources_pair in enumerate(resources or []):
    type, path = resources_pair
    parameters["file_resources-%d-type" % i] = type
    parameters["file_resources-%d-path" % i] = path
    parameters["file_resources-%d-_exists" % i] = 'True'
  response = client.post("/beeswax/execute", parameters,
    follow=follow)

  if wait:
    return wait_for_query_to_finish(client, response)
  return response


def verify_history(client, fragment, design=None, reverse=False):
  """
  Verify that the query fragment and/or design are in the query history.
  If reverse is True, verify the opposite.
  Return the size of the history.
  """
  resp = client.get('/beeswax/query_history')
  my_assert = reverse and assert_false or assert_true
  my_assert(fragment in resp.content)
  if design:
    my_assert(design in resp.content)
  return len(resp.context['page'].object_list)


class BeeswaxSampleProvider(object):
  """
  Setup the test db and install sample data
  """
  @classmethod
  def setup_class(cls):
    cls.cluster, shutdown = get_shared_beeswax_server()
    cls.client = make_logged_in_client()
    # Weird redirection to avoid binding nonsense.
    cls.shutdown = [ shutdown ]
    cls.init_beeswax_db()

  @classmethod
  def teardown_class(cls):
    cls.shutdown[0]()

  @classmethod
  def init_beeswax_db(cls):
    """
    Install the common test tables (only once)
    """
    global _INITIALIZED
    if _INITIALIZED:
      return

    # Create a test table and load data here;
    # this is used in several tests.
    CREATE_TABLE = """
      CREATE TABLE test (foo INT, bar STRING)
      ROW FORMAT DELIMITED
        FIELDS TERMINATED BY '\t'
        LINES TERMINATED BY '\n'
    """
    make_query(cls.client, CREATE_TABLE, wait=True)

    # Create some data for it
    cls.cluster.fs.setuser(cls.cluster.superuser)
    def write_sample_data(cluster, filename):
      f = cluster.fs.open(filename, "w")
      for x in range(256):
        f.write("%d\t0x%x\n" % (x, x))
      f.close()
    write_sample_data(cls.cluster, "/tmp/sample_data.tsv")
    write_sample_data(cls.cluster, "/tmp/sample_data2.tsv")

    # Load the data
    LOAD_DATA = """
      LOAD DATA INPATH '/tmp/sample_data.tsv' OVERWRITE INTO TABLE test
    """
    make_query(cls.client, LOAD_DATA, wait=True, local=False)

    CREATE_PARTITIONED_TABLE = """
      CREATE TABLE test_partitions (foo INT, bar STRING)
      PARTITIONED BY (baz STRING, boom STRING)
      ROW FORMAT DELIMITED
        FIELDS TERMINATED BY '\t'
        LINES TERMINATED BY '\n'
    """
    make_query(cls.client, CREATE_PARTITIONED_TABLE, wait=True)

    LOAD_DATA = """
      LOAD DATA INPATH '/tmp/sample_data2.tsv' OVERWRITE INTO TABLE test_partitions PARTITION (baz='baz_one', boom='boom_two')
    """
    make_query(cls.client, LOAD_DATA, wait=True, local=False)
    _INITIALIZED = True
